# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
sMRIPrep's top-level module.

API for developers
==================


"""
from .__about__ import (
    __copyright__,
    __credits__,
    __version__,
)

__all__ = [
    '__copyright__',
    '__credits__',
    '__version__',
]
