# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

# Load modules for compatibility
from niworkflows.interfaces.bids import DerivativesDataSink as DDS


class DerivativesDataSink(DDS):
    out_path_base = 'smriprep'


del DDS
